/*
 *  linux/drivers/block/vroot.c
 *
 *  Written by Herbert Ptzl, 9/11/2002
 *
 *  based on the loop.c code by Theodore Ts'o.
 *
 * Copyright 2002-2003 by Herbert Ptzl.
 * Redistribution of this file is permitted under the
 * GNU General Public License.
 *
 */

#define MAJOR_NR VROOT_MAJOR

#include <linux/config.h>
#include <linux/module.h>

#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/stat.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/wait.h>
#include <linux/blk.h>
#include <linux/blkpg.h>
#include <linux/init.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/smp_lock.h>
#include <linux/swap.h>
#include <linux/slab.h>

#include <asm/uaccess.h>

#include "vroot.h"

static int max_vroot = MAX_VROOT_DEFAULT;
static struct vroot_device *vroot_dev;
static devfs_handle_t devfs_handle;      /*  For the directory */

#ifdef	MODULE
typedef	kdev_t	(*vroot_get_dev_f)(int dev);

extern int register_vroot_get_dev(vroot_get_dev_f);
extern int unregister_vroot_get_dev(vroot_get_dev_f);

static	kdev_t _vroot_get_dev(int dev)
#else
	kdev_t vroot_get_dev(int dev)
#endif
{
	struct vroot_device *vr;

	if (dev >= max_vroot)
		return NODEV;

	vr = &vroot_dev[dev];
	if (vr->vr_state != Vr_bound)
		return NODEV;
	dprintk(KERN_INFO "vroot[%d]_get_dev: dev(%d,%d)\n",
		dev, MAJOR(vr->vr_device), MINOR(vr->vr_device));
	return vr->vr_device;
}

static int vroot_set_dev(
	struct vroot_device *vr,
	struct file *vr_file,
	kdev_t dev,
	unsigned int arg)
{
	struct file	*file;
	struct inode	*inode;
	int		error;

	MOD_INC_USE_COUNT;

	error = -EBUSY;
	if (vr->vr_state != Vr_unbound)
		goto out;

	error = -EBADF;
	file = fget(arg);
	if (!file)
		goto out;
		
	error = -EINVAL;
	inode = file->f_dentry->d_inode;

	if (S_ISBLK(inode->i_mode)) {
		vr->vr_device = inode->i_rdev;
		if (vr->vr_device == dev) {
			error = -EBUSY;
			goto out_fput;
		}
	} else
		goto out_fput;
		
	dprintk(KERN_INFO "vroot[%d]_set_dev: dev(%d,%d)\n",
		vr->vr_number,
		MAJOR(inode->i_rdev), MINOR(inode->i_rdev));

	vr->vr_state = Vr_bound;
	fput(file);
	return 0;
 out_fput:
	fput(file);
 out:
	MOD_DEC_USE_COUNT;
	return error;
}

static int vroot_clr_dev(
	struct vroot_device *vr,
	struct file *vr_file,
	kdev_t dev)
{
	if (vr->vr_state != Vr_bound)
		return -ENXIO;
	if (vr->vr_refcnt > 1)	/* we needed one fd for the ioctl */
		return -EBUSY;

	dprintk(KERN_INFO "vroot[%d]_clr_dev: dev(%d,%d)\n",
		vr->vr_number,
		MAJOR(vr->vr_device), MINOR(vr->vr_device));

	vr->vr_state = Vr_unbound;
	vr->vr_device = NODEV;
	MOD_DEC_USE_COUNT;
	return 0;
}

static int vroot_make_request(
	request_queue_t *q,
	int rw,
	struct buffer_head *rbh)
{
	if (!buffer_locked(rbh))
		BUG();

	if (MINOR(rbh->b_rdev) >= max_vroot)
		goto out;

	dprintk(KERN_WARNING "vroot[%d]_make_request: denied.\n",
		MINOR(rbh->b_rdev));	
 out:
	buffer_IO_error(rbh);
	return 0;
}

static int vr_ioctl(
	struct inode * inode,
	struct file * file,
	unsigned int cmd,
	unsigned long arg)
{
	struct vroot_device *vr;
	int dev, err;

	if (!inode)
		return -EINVAL;
	if (MAJOR(inode->i_rdev) != MAJOR_NR) {
		dprintk(KERN_WARNING "vr_ioctl: pseudo-major != %d\n",
		       MAJOR_NR);
		return -ENODEV;
	}
	dev = MINOR(inode->i_rdev);
	if (dev >= max_vroot)
		return -ENODEV;
	vr = &vroot_dev[dev];
	down(&vr->vr_ctl_mutex);
	switch (cmd) {
	case VROOT_SET_DEV:
		err = vroot_set_dev(vr, file, inode->i_rdev, arg);
		break;
	case VROOT_CLR_DEV:
		err = vroot_clr_dev(vr, file, inode->i_rdev);
		break;
	default:
		err = -EINVAL;
		break;
	}
	up(&vr->vr_ctl_mutex);
	return err;
}

static int vr_open(
	struct inode *inode,
	struct file *file)
{
	struct vroot_device *vr;
	int	dev;

	if (!inode)
		return -EINVAL;
	if (MAJOR(inode->i_rdev) != MAJOR_NR) {
		dprintk(KERN_WARNING "vr_open: pseudo-major != %d\n", MAJOR_NR);
		return -ENODEV;
	}
	dev = MINOR(inode->i_rdev);
	if (dev >= max_vroot)
		return -ENODEV;

	vr = &vroot_dev[dev];
	MOD_INC_USE_COUNT;
	down(&vr->vr_ctl_mutex);

	vr->vr_refcnt++;
	up(&vr->vr_ctl_mutex);
	return 0;
}

static int vr_release(
	struct inode *inode,
	struct file *file)
{
	struct vroot_device *vr;
	int	dev;

	if (!inode)
		return 0;
	if (MAJOR(inode->i_rdev) != MAJOR_NR) {
		dprintk(KERN_WARNING "vr_release: pseudo-major != %d\n",
		       MAJOR_NR);
		return 0;
	}
	dev = MINOR(inode->i_rdev);
	if (dev >= max_vroot)
		return 0;

	vr = &vroot_dev[dev];
	down(&vr->vr_ctl_mutex);

	vr->vr_refcnt--;
	up(&vr->vr_ctl_mutex);
	MOD_DEC_USE_COUNT;
	return 0;
}

static struct block_device_operations vr_fops = {
	owner:		THIS_MODULE,
	open:		vr_open,
	release:	vr_release,
	ioctl:		vr_ioctl,
};

/*
 * And now the modules code and kernel interface.
 */
MODULE_PARM(max_vroot, "i");
MODULE_PARM_DESC(max_vroot, "Maximum number of vroot devices (1-256)");
MODULE_LICENSE("GPL");

MODULE_AUTHOR ("Herbert Ptzl");
MODULE_DESCRIPTION ("Virtual Root Device Mapper");


int __init vroot_init(void)
{
	int	i;

	if ((max_vroot < 1) || (max_vroot > 256)) {
		printk(KERN_WARNING "vroot: invalid max_vroot (must be between"
				    " 1 and 256), using default (%d)\n",
				    MAX_VROOT_DEFAULT);
		max_vroot = MAX_VROOT_DEFAULT;
	}

	if (devfs_register_blkdev(MAJOR_NR, "vroot", &vr_fops)) {
		printk(KERN_WARNING "Unable to get major number %d for vroot"
				    " device\n", MAJOR_NR);
		return -EIO;
	}

	devfs_handle = devfs_mk_dir(NULL, "vroot", NULL);
	devfs_register_series(devfs_handle, "%u", max_vroot,
			      DEVFS_FL_DEFAULT, MAJOR_NR, 0,
			      S_IFBLK | S_IRUSR | S_IWUSR | S_IRGRP,
			      &vr_fops, NULL);

	vroot_dev = kmalloc(max_vroot * sizeof(struct vroot_device), GFP_KERNEL);
	if (!vroot_dev)
		return -ENOMEM;

	blk_queue_make_request(BLK_DEFAULT_QUEUE(MAJOR_NR), vroot_make_request);

	for (i = 0; i < max_vroot; i++) {
		struct vroot_device *vr = &vroot_dev[i];
		memset(vr, 0, sizeof(struct vroot_device));
		init_MUTEX(&vr->vr_ctl_mutex);
		vr->vr_number = i;
		vr->vr_state = Vr_unbound;
	}

	for (i = 0; i < max_vroot; i++)
		register_disk(NULL, MKDEV(MAJOR_NR, i), 1, &vr_fops, 0);

#ifdef	MODULE
	register_vroot_get_dev(_vroot_get_dev);
#endif
	printk(KERN_INFO "vroot: loaded (max %d devices)\n", max_vroot);
	return 0;
}

void vroot_exit(void)
{
#ifdef	MODULE
	unregister_vroot_get_dev(_vroot_get_dev);
#endif	
	devfs_unregister(devfs_handle);
	if (devfs_unregister_blkdev(MAJOR_NR, "vroot"))
		printk(KERN_WARNING "vroot: cannot unregister blkdev\n");

	kfree(vroot_dev);
}

module_init(vroot_init);
module_exit(vroot_exit);

#ifndef MODULE
static int __init max_vroot_setup(char *str)
{
	max_vroot = simple_strtol(str, NULL, 0);
	return 1;
}

__setup("max_vroot=", max_vroot_setup);

#endif
