/*
 *  linux/kernel/vcontext.c
 *
 *  Virtual Context Support
 *
 *  Copyright (C) 2003  Herbert Ptzl
 *
 *  V0.01  context helper
 *
 */

#include <linux/config.h>
#include <linux/linkage.h>
#include <linux/utsname.h>
#include <linux/slab.h>
#include <linux/vcontext.h>
#include <linux/vswitch.h>
#include <linux/sched.h>
#include <linux/kernel_stat.h>

#include <asm/errno.h>
#include <asm/uaccess.h>


int vc_ctx_kill(uint32_t id, void *data)
{
	int retval, count=0;
	struct vcmd_ctx_kill_v0 vc_data;
	struct siginfo info;
	struct task_struct *p;
	pid_t initpid = 0;

	if (copy_from_user (&vc_data, data, sizeof(vc_data)))
		return -EFAULT;
	if (!vx_check(0, VX_ADMIN))
		return -EPERM;
	
	info.si_signo = vc_data.sig;
	info.si_errno = 0;
	info.si_code = SI_USER;
	info.si_pid = current->pid;
	info.si_uid = current->uid;

	retval = -ESRCH;
	read_lock(&tasklist_lock);
	switch (vc_data.pid) {
	case -1:
	case  0:
		for_each_task(p) {
			if (!initpid && p->vx_id == id && p->s_info)
				initpid = p->s_info->initpid;
			if (p->vx_id == id  && p->pid > 1
				&& (!vc_data.pid || initpid != p->pid)
				&& thread_group_leader(p)) {
				int err = send_sig_info(vc_data.sig, &info, p);

				++count;
				if (err != -EPERM)
					retval = err;
			}
		}
		break;
		
	default:
	p = find_task_by_pid(vc_data.pid);
		if (p) {
			if (!thread_group_leader(p)) {
				struct task_struct *tg;
			
				tg = find_task_by_pid(p->tgid);
				if (tg)
					p = tg;
			}
			if ((id == -1) || (p->vx_id == id))
				retval = send_sig_info(vc_data.sig, &info, p);
		}
		break;
	}
	read_unlock(&tasklist_lock);
	return retval;
}

int vc_get_rlimit(uint32_t id, void *data)
{
	return -ENOSYS;
}

int vc_set_rlimit(uint32_t id, void *data)
{
	return -ENOSYS;
}

int vc_get_rlimit_mask(uint32_t id, void *data)
{
	return -ENOSYS;
}



/*  system functions */


/*
 *	Alloc a new s_info to the current process and release
 *	the one currently owned by the current process.
 */
static void vx_alloc_info(void)
{
	struct context_info *s_info;

	s_info = kmalloc(sizeof(struct context_info), GFP_KERNEL);
	if (s_info) {
		int i;
		memset (s_info, 0, sizeof(*s_info));
		s_info->vx_id[0] = current->vx_id;
		atomic_set(&s_info->refcount, 1);
		atomic_set(&s_info->ticks, current->counter);
		s_info->flags = 0;
		s_info->initpid = 0;
		s_info->nr_threads = 1;
		s_info->total_forks = 0;
		s_info->bias_cswtch = kstat.context_swtch;
		s_info->bias_jiffies = jiffies;
		s_info->bias_idle = init_tasks[0]->times.tms_utime +
			init_tasks[0]->times.tms_stime;
		for (i=0; i<RLIM_NLIMITS; i++) {
			s_info->rlim[i] = 0xffffffff;
			atomic_set(&s_info->res[i], 0);
		}
		down_read(&uts_sem);
		if (current->s_info) {
			strcpy(s_info->nodename, current->s_info->nodename);
			strcpy(s_info->domainname, current->s_info->domainname);
		} else {
			strcpy(s_info->nodename, system_utsname.nodename);
			strcpy(s_info->domainname, system_utsname.domainname);
		}
		up_read(&uts_sem);
		vx_release_info(current);	
		current->s_info = s_info;
		/*
			The current process is switching to a new context
			so we preset the open file counter with
			the file currently open by that process.
			Some of those files may have been opened by
			a parent, so do not strictly belong to this
			process, so we kind of over bill the current process
			but it is minimal.
		*/
		atomic_set(&s_info->res[RLIMIT_NOFILE],
			atomic_read(&current->files->count));
	}
}

rwlock_t ctx_ref_lock = RW_LOCK_UNLOCKED;

/*
 *	Increase the reference count on the context_info member of a task
 */
void vx_assign_info (struct task_struct *p)
{
	read_lock(&ctx_ref_lock);
	if (p->s_info)
		atomic_inc(&p->s_info->refcount);
	read_unlock(&ctx_ref_lock);
}

/*
 *	Decrease the reference count on the context_info member of a task
 *	Free the struct if the reference count reach 0.
 */
void vx_release_info (struct task_struct *p)
{
	write_lock(&ctx_ref_lock);
	if (p->s_info) {
		if (atomic_dec_and_test(&p->s_info->refcount)) {
			struct context_info *s_info;

			s_info = p->s_info;
			p->s_info = NULL;
			kfree(s_info);
		}
	}
	write_unlock(&ctx_ref_lock);
}

/*
 *	Alloc a new ip_info to the current process and release
 *	the one currently owned by the current process.
 */
static void vx_alloc_ip_info(void)
{
	struct iproot_info *ip_info =
		kmalloc(sizeof(struct iproot_info), GFP_KERNEL);

	memset(ip_info, 0, sizeof(*ip_info));
	ip_info->mark = 0xdeadbeef;
	atomic_set(&ip_info->refcount, 1);
	vx_release_ip_info(current->ip_info);	
	current->ip_info = ip_info;
}

/*
 *	Increase the reference count on the ip_info member of a task
 */
void vx_assign_ip_info (struct iproot_info *ip_info)
{
	if (ip_info) {
		atomic_inc(&ip_info->refcount);
		if (ip_info->mark != 0xdeadbeef)
			printk("vx_assign_ip_info: broken signature %08lx\n", ip_info->mark);
	}
}

/*
 *	Decrease the reference count on the ip_info struct
 *	Free the struct if the reference count reach 0.
 */
void vx_release_ip_info (struct iproot_info *ip_info)
{
	if (ip_info)
		if (atomic_dec_and_test(&ip_info->refcount))
			kfree(ip_info);
}


static int vx_switch_user_struct(int new_context)
{
	struct user_struct *new_user;

	new_user = alloc_uid(new_context, current->uid);
	if (!new_user)
		return -ENOMEM;

	if (new_user != current->user) {
		struct user_struct *old_user = current->user;

		atomic_inc(&new_user->processes);
		atomic_dec(&old_user->processes);
		current->user = new_user;
		free_uid(old_user);
	} else
		free_uid(new_user);
	return 0;
}

static int vx_set_initpid(int flags)
{
	int ret = 0;
	if (flags & VX_INFO_INIT) {
		if (current->s_info == NULL)
			ret = -EINVAL;
		else if (current->s_info->initpid != 0)
			ret = -EPERM;
		else
			current->s_info->initpid = current->tgid;
	}
	return ret;
}


/*  new security context (syscall) */

/*
 *	Change to a new security context and reduce the capability
 *	basic set of the current process
 */
int vc_new_s_context(uint32_t ctx, void *data)
{
	int ret = -EPERM;
	struct vcmd_new_s_context_v1 vc_data;

	if (copy_from_user(&vc_data, data, sizeof(vc_data)))
		return -EFAULT;
	if (ctx == -1) {
		if (current->s_info == NULL
			|| !(current->s_info->flags & VX_INFO_LOCK)) {
			/* Ok we allocate a new context. For now, we just increase */
			/* it. Wrap around possible, so we loop */
			static int new_xid = MAX_S_CONTEXT;
			static spinlock_t alloc_ctx_lock = SPIN_LOCK_UNLOCKED;
			int old_xid = current->vx_id;
			int barrier = new_xid;
			int valid = 0;

			spin_lock(&alloc_ctx_lock);
			do {
				struct task_struct *p;
				
				valid = 1;
				if (++new_xid > MAX_S_CONTEXT)
					new_xid = MIN_D_CONTEXT;

				/* Check if in use */
				read_lock(&tasklist_lock);
				for_each_task(p) {
					if (p->vx_id == new_xid) {
						valid = 0;
						break;
					}
				}
				read_unlock(&tasklist_lock);
				
				if (valid) {
					current->vx_id = new_xid;
					break;
				}
			} while (barrier != new_xid);
			spin_unlock(&alloc_ctx_lock);
			
			if (!valid)
				return -EDEADLK;
			
			ret = vx_switch_user_struct(new_xid);
			if (ret == 0) {
				current->cap_bset &= (~vc_data.remove_cap);
				ret = new_xid;
				vx_alloc_info();
				if (current->s_info) {
					vx_set_initpid(vc_data.flags);
					current->s_info->flags |= vc_data.flags;
				}
			} else
				current->vx_id = old_xid;
		}
	} else if (ctx == -2) {
		ret = vx_set_initpid(vc_data.flags);
		if (ret == 0) {
			/* We keep the same vx_id, but lower the capabilities */
			current->cap_bset &= (~vc_data.remove_cap);
			ret = current->vx_id;
			if (current->s_info) {
				if (vc_data.flags & VX_INFO_INIT)
					current->s_info->initpid = current->tgid;
				current->s_info->flags |= vc_data.flags;
			}
		}
	} else if (ctx <= 0 || ctx > MAX_S_CONTEXT) {
		ret = -EINVAL;
	} else if (vx_check(0, VX_ADMIN)
		&& capable(CAP_SYS_ADMIN)
		&& (current->s_info == NULL
			||(current->s_info->flags & VX_INFO_LOCK) == 0)) {
		/* The root context can become any context it wants */
		int found = 0;
		struct task_struct *p;

		/* Check if in use so we reuse the same context_info */
		read_lock(&tasklist_lock);
		ret = ctx;
		for_each_task(p) {
			if (p->vx_id == ctx) {
				found = 1;
				if (p->s_info == NULL
					|| !(p->s_info->flags & VX_INFO_PRIVATE)) {
					vx_release_info(current);
					vx_assign_info(p);
					current->s_info = p->s_info;
				}
				else
					ret = -EPERM;
				break;
			}
		}
		read_unlock(&tasklist_lock);
		if (ret == ctx) {
			ret = vx_switch_user_struct(ctx);
			if (ret == 0) {
				current->vx_id = ctx;
				current->cap_bset &= (~vc_data.remove_cap);
				if (!found)
					vx_alloc_info();
				if (current->s_info)
					current->s_info->flags |= vc_data.flags;
			}
		}
	}
	return ret;
}


/*  set ipv4 root (syscall) */

int vc_set_ipv4root(uint32_t nbip, void *data)
{
	int ret = -EPERM;
	struct vcmd_set_ipv4root_v3 vc_data;
	struct iproot_info *ip_info = current->ip_info;

	if (copy_from_user (&vc_data, data, sizeof(vc_data)))
		return -EFAULT;

	if (nbip < 0 || nbip > NB_IPV4ROOT)
		ret = -EINVAL;
	if (!ip_info || ip_info->ipv4[0] == 0 || capable(CAP_NET_ADMIN))
		// We are allowed to change everything
		ret = 0;
	else if (current->ip_info) {
		// We are allowed to select a subset of the currently
		// installed IP numbers. No new one allowed
		// We can't change the broadcast address though
		int i;
		int found = 0;
		for (i=0; i<nbip; i++) {
			int j;
			__u32 ipi = vc_data.ip_mask_pair[i].ip;
			for (j=0; j<ip_info->nbipv4; j++) {
				if (ipi == ip_info->ipv4[j]) {
					found++;
					break;
				}
			}
		}
		if (found == nbip && vc_data.broadcast == ip_info->v4_bcast)
			ret = 0;
	}
	if (ret == 0) {
		int i;

		vx_alloc_ip_info(); /* release existing? */
		ip_info = current->ip_info;
		ip_info->nbipv4 = nbip;
		for (i=0; i<nbip; i++) {
			ip_info->ipv4[i] = vc_data.ip_mask_pair[i].ip;
			ip_info->mask[i] = vc_data.ip_mask_pair[i].mask;
		}
		ip_info->v4_bcast = vc_data.broadcast;
	}
	return ret;
}


#include <linux/proc_fs.h>

int vx_proc_ioctl(struct inode * inode, struct file * filp,
	unsigned int cmd, unsigned long arg)
{
	struct proc_dir_entry *entry;
	int error = 0;
	int flags;

	if (inode->i_ino < PROC_DYNAMIC_FIRST ||
		inode->i_ino >= PROC_DYNAMIC_FIRST+PROC_NDYNAMIC)
		return -ENOTTY;

	entry = (struct proc_dir_entry *) inode->u.generic_ip;

	switch(cmd) {
	case FIOC_GETXFLG: {
		/* fixme: if stealth, return -ENOTTY */
		error = -EPERM;
		flags = entry->vx_flags;
		if (capable(CAP_CONTEXT))
			error = put_user(flags, (int *) arg);
		break;
	}
	case FIOC_SETXFLG: {
		/* fixme: if stealth, return -ENOTTY */
		error = -EPERM;
		if (!capable(CAP_CONTEXT))
			break;
		error = -EROFS;
		if (IS_RDONLY(inode))
			break;
		error = -EFAULT;
		if (get_user(flags, (int *) arg))
			break;
		error = 0;
		entry->vx_flags = flags;
		break;
	}
	default:
		return -ENOTTY;
	}
	return error;
}

